#include <iostream>
#include <algorithm>
#include <math.h>

#include "gtest/gtest.h"

#include <cdf.h>

#include "images.h"

using namespace std;


// These tests rely upon the input image being generated from the sequence
// image[i] = i, where i is the index from 1 to the number of pixels.

// index in sequence sum(0,n) i
// where sum(0,i)i / sum(0,n) i == fraction
long pdf_clip_idx( long npix, double fraction ) {

    return ( sqrt( ( 4 * npix * npix + 4 * npix) * fraction + 1) - 1 )/ 2;

}

long psum( long n ) { return n * (n + 1) / 2;  }

struct CmpCDF {

    CDF& cdf;
    long int npix;
    long int idx;
    long int offset;
    double max;

    CmpCDF( CDF& cdf, long int npix, long int idx = 0, long int offset = 0 ) :
	cdf(cdf), npix(npix), idx(idx), offset(offset), max( psum(npix) + offset )
    {};

    void operator() (CDF::Pixel& pixel ) {

	idx++;
	ASSERT_EQ( ( psum(idx) + offset ) / max, pixel.p );
    }

};

struct CmpImage {

    CDF& cdf;
    long int idx;

    CmpImage( CDF& cdf ) : cdf(cdf), idx(0) { };

    void operator() (CDF::Pixel& pixel ) {

	idx++;
	ASSERT_EQ( idx, pixel.p ) ;
    }

};

TEST( read, image ) {

    CDF cdf;

    long extent[2];
    string filename( "sequence.fits" );

    cdf.read_image( filename );

    get_image_info( filename, extent );

    ASSERT_EQ( extent[0] * extent[1], cdf.pixels.size() );
    for_each( cdf.pixels.begin(), cdf.pixels.end(), CmpImage( cdf) );


    //    ostream_iterator<CDF::Pixel>err ( std::cerr, "\n" );
    //    copy( cdf.pixels.begin(), cdf.pixels.end(), err );

}

TEST( read, normalize ) {

    CDF cdf( "sequence.fits" );
    long extent[2];
    string filename( "sequence.fits" );

    get_image_info( filename, extent );

    ASSERT_EQ( extent[0] * extent[1], cdf.pixels.size() );
    for_each( cdf.pixels.begin(), cdf.pixels.end(), CmpCDF( cdf, cdf.pixels.size() ) );
}


TEST( read, cdf_clip_lo ) {

    CDF cdf( "sequence.fits", 0, 0, 1, 0.5, 1, "cdf"  );
    long extent[2];
    string filename( "sequence.fits" );

    get_image_info( filename, extent );
    double npix = extent[0] * extent[1];

    double lo = pdf_clip_idx( npix, 0.5 );

    ASSERT_EQ( npix - lo, cdf.pixels.size() );
    for_each( cdf.pixels.begin(), cdf.pixels.end(), CmpCDF( cdf, npix, lo, -psum(lo) ) );

}

TEST( read, cdf_clip_hi ) {

    CDF cdf( "sequence.fits", 0, 0, 1, 0, .5, "cdf"  );
    long extent[2];
    string filename( "sequence.fits" );

    get_image_info( filename, extent );
    double npix = extent[0] * extent[1];

    double hi = pdf_clip_idx( npix, 0.5 );

    ASSERT_EQ( hi, cdf.pixels.size() );
    for_each( cdf.pixels.begin(), cdf.pixels.end(), CmpCDF( cdf, cdf.pixels.size() ) );

}

TEST( read, pdf_clip_lo ) {

    long extent[2];
    string filename( "sequence.fits" );

    get_image_info( filename, extent );
    double npix = extent[0] * extent[1];
    double max = psum(npix);

    // choose fraction s.t. reject half of pixels
    double fraction = npix / 2 / max;

    CDF cdf( "sequence.fits", 0, 0, 1, fraction, 1, "pdf" );

    ASSERT_EQ( npix/2, cdf.pixels.size() );
    for_each( cdf.pixels.begin(), cdf.pixels.end(), CmpCDF( cdf, npix, npix/2, -psum(npix/2) ) );

}

TEST( read, pdf_clip_hi ) {

    long extent[2];
    string filename( "sequence.fits" );

    get_image_info( filename, extent );
    double npix = extent[0] * extent[1];
    double max = psum(npix);

    // choose fraction s.t. reject half of pixels
    double fraction = npix / 2 / max;

    CDF cdf( "sequence.fits", 0, 0, 1, 0, fraction, "pdf" );

    ASSERT_EQ( npix/2, cdf.pixels.size() );
    for_each( cdf.pixels.begin(), cdf.pixels.end(), CmpCDF( cdf, cdf.pixels.size()) );

}
